"""Standard authentication backend."""

from __future__ import annotations

import logging
from typing import Any, Dict, Optional, Set, TYPE_CHECKING, Union, cast

from django.conf import settings
from django.contrib.auth import hashers
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import User
from django.utils.translation import gettext_lazy as _
from djblets.db.query import get_object_or_none

from reviewboard.accounts.backends.base import BaseAuthBackend
from reviewboard.accounts.forms.auth import StandardAuthSettingsForm
from reviewboard.accounts.models import LocalSiteProfile
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractBaseUser, AnonymousUser
    from django.db.models import Model
    from django.http import HttpRequest


logger = logging.getLogger(__name__)


class StandardAuthBackend(BaseAuthBackend, ModelBackend):
    """Authenticate users against the local database.

    This will authenticate a user against their entry in the database, if
    the user has a local password stored. This is the default form of
    authentication in Review Board.

    This backend also handles permission checking for users on LocalSites.
    In Django, this is the responsibility of at least one auth backend in
    the list of configured backends.

    Regardless of the specific type of authentication chosen for the
    installation, StandardAuthBackend will always be provided in the list
    of configured backends. Because of this, it will always be able to
    handle authentication against locally added users and handle
    LocalSite-based permissions for all configurations.
    """

    backend_id = 'builtin'
    name = _('Standard Registration')
    settings_form = StandardAuthSettingsForm
    supports_registration = True
    supports_change_name = True
    supports_change_email = True
    supports_change_password = True

    _VALID_LOCAL_SITE_PERMISSIONS = [
        'hostingsvcs.change_hostingserviceaccount',
        'hostingsvcs.create_hostingserviceaccount',
        'reviews.add_group',
        'reviews.can_change_status',
        'reviews.can_edit_reviewrequest',
        'reviews.can_submit_as_another_user',
        'reviews.can_view_invite_only_groups',
        'reviews.change_default_reviewer',
        'reviews.change_group',
        'reviews.delete_file',
        'reviews.delete_screenshot',
        'scmtools.add_repository',
        'scmtools.change_repository',
    ]

    def authenticate(
        self,
        request: Optional[HttpRequest] = None,
        *,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs,
    ) -> Optional[User]:
        """Authenticate the user.

        This will attempt to authenticate the user against the database.
        If the username and password are valid, a user will be returned.

        Version Changed:
            6.0:
            * ``request`` is now optional.
            * ``username`` and ``password`` are technically optional, to
              aid in consistency for type hints, but will result in a ``None``
              result.

        Version Changed:
            4.0:
            The ``request`` argument is now mandatory as the first positional
            argument, as per requirements in Django.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the caller. This may be ``None``.

            username (str):
                The username used for authentication.

            password (str):
                The password used for authentication.

            **kwargs (dict, unused):
                Additional keyword arguments supplied by the caller.

        Returns:
            django.contrib.auth.models.User:
            The authenticated user, or ``None`` if the user could not be
            authenticated for any reason.
        """
        if not username or not password:
            # This may be an authentication request for a backend expecting
            # different arguments.
            return None

        return cast(
            User,
            ModelBackend.authenticate(self,
                                      request,
                                      username=username,
                                      password=password,
                                      **kwargs))

    def get_or_create_user(
        self,
        username: str,
        request: Optional[HttpRequest] = None,
    ) -> Optional[User]:
        """Return an existing user or create one if it doesn't exist.

        This does not authenticate the user.

        Args:
            username (str):
                The username to fetch or create.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            django.contrib.auth.models.User:
            The resulting user, or ``None`` if one could not be found.
        """
        return get_object_or_none(User, username=username)

    def update_password(
        self,
        user: User,
        password: str,
    ) -> None:
        """Update a user's password on the backend.

        This will update the user information, but will not save the user
        instance. That must be saved manually.

        Args:
            user (django.contrib.auth.models.User):
                The user whose password will be changed.

            password (str):
                The new password.
        """
        user.password = hashers.make_password(password)

    def get_all_permissions(
        self,
        user: Union[AbstractBaseUser, AnonymousUser],
        obj: Optional[Model] = None,
    ) -> Set[str]:
        """Return a list of all permissions for a user.

        If a LocalSite instance is passed as ``obj``, then the permissions
        returned will be those that the user has on that LocalSite. Otherwise,
        they will be their global permissions.

        It is not legal to pass any other object.

        Args:
            user (django.contrib.auth.models.AnonymousUser or
                  django.contrib.auth.models.User):
                The user to retrieve permission for.

            obj (django.db.models.Model, optional):
                A model used as context.

                If provided, this must be a
                :py:class:`~reviewboard.site.models.LocalSite`.

        Returns:
            set of str:
            A set of all permission codes.
        """
        if obj is not None and not isinstance(obj, LocalSite):
            logger.error('Unexpected object %r passed to '
                         'StandardAuthBackend.get_all_permissions. '
                         'Returning an empty list.',
                         obj)

            if settings.DEBUG:
                raise ValueError('Unexpected object %r' % obj)

            return set()

        if user.is_anonymous:
            return set()

        # First, get the list of all global permissions.
        #
        # Django's ModelBackend doesn't support passing an object, and will
        # return an empty set, so don't pass an object for this attempt.
        permissions = \
            super(StandardAuthBackend, self).get_all_permissions(user)

        if obj is not None:
            # We know now that this is a LocalSite, due to the assertion
            # above.
            local_site_perm_cache: Dict[str, Set[str]]

            try:
                local_site_perm_cache = getattr(user, '_local_site_perm_cache')
            except AttributeError:
                local_site_perm_cache = {}
                setattr(user, '_local_site_perm_cache', local_site_perm_cache)

            if obj.pk not in local_site_perm_cache:
                perm_cache: Set[str] = set()

                try:
                    site_profile = user.get_site_profile(
                        obj,
                        create_if_missing=False)
                    site_perms = site_profile.permissions or {}

                    if site_perms:
                        perm_cache = {
                            key
                            for key, value in site_perms.items()
                            if value
                        }
                except LocalSiteProfile.DoesNotExist:
                    pass

                local_site_perm_cache[obj.pk] = perm_cache

            permissions = permissions.copy()
            permissions.update(local_site_perm_cache[obj.pk])

        return permissions

    def has_perm(
        self,
        user: Union[AbstractBaseUser, AnonymousUser],
        perm: str,
        obj: Any = None,
    ) -> bool:
        """Return whether or not a user has the given permission.

        If a LocalSite instance is passed as ``obj``, then the permissions
        checked will be those that the user has on that LocalSite. Otherwise,
        they will be their global permissions.

        It is not legal to pass any other object.

        Args:
            user (django.contrib.auth.models.AnonymousUser or
                  django.contrib.auth.models.User):
                The user to retrieve permission for.

            perm (str):
                The permission code to check for.

            obj (django.db.models.Model, optional):
                A model used as context.

                If provided, this must be a
                :py:class:`~reviewboard.site.models.LocalSite`.

        Returns:
            bool:
            ``True`` if the user has the permission. ``False`` if it does not.
        """
        if obj is not None and not isinstance(obj, LocalSite):
            logger.error('Unexpected object %r passed to has_perm. '
                         'Returning False.', obj)

            if settings.DEBUG:
                raise ValueError('Unexpected object %r' % obj)

            return False

        if not user.is_active:
            return False

        if obj is not None:
            local_site_admin_for: Dict[str, bool]

            try:
                local_site_admin_for = getattr(user, '_local_site_admin_for')
            except AttributeError:
                local_site_admin_for = {}
                setattr(user, '_local_site_admin_for', local_site_admin_for)

            try:
                is_mutable = local_site_admin_for[obj.pk]
            except KeyError:
                is_mutable = obj.is_mutable_by(user)
                local_site_admin_for[obj.pk] = is_mutable

            if is_mutable:
                return perm in self._VALID_LOCAL_SITE_PERMISSIONS

        return super().has_perm(user, perm, obj)



"""Utilities for WebHook handlers."""

from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from django.conf import settings
from django.db.models import Q
from django.shortcuts import get_object_or_404
from housekeeping import deprecate_non_keyword_only_args

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.reviews.models import ReviewRequest
from reviewboard.scmtools.models import Repository
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


logger = logging.getLogger(__name__)


def get_git_branch_name(
    ref_name: str,
) -> str:
    """Return the branch name corresponding to the specified ref name.

    Args:
        ref_name (str):
            The name of the Git reference.

    Returns:
        str:
        The isolated branch name.
    """
    if ref_name.startswith('refs/heads/'):
        return ref_name[len('refs/heads/'):]
    else:
        return ref_name


@deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
def get_repository_for_hook(
    *,
    repository_id: int,
    hosting_service_id: str,
    local_site_name: str | None,
    hooks_uuid: (str | None) = None,
) -> Repository:
    """Return a Repository for the given hook parameters.

    Version Changed:
        8.0:
        Made arguments keyword-only.

    Args:
        repository_id (int):
            The ID of the repository.

        hosting_service_id (str):
            The name of the hosting service.

        local_site_name (str):
            The Local Site name, if available.

        hooks_uuid (str, optional):
            The UUID to use for WebHooks for this repository.

    Returns:
        reviewboard.scmtools.models.Repository:
        The repository object.

    Raises:
        django.http.Http404:
            A repository with the given parameters was not found.
    """
    q = (Q(pk=repository_id) &
         Q(hosting_account__service_name=hosting_service_id))

    if hooks_uuid:
        q &= Q(hooks_uuid=hooks_uuid)

    if local_site_name:
        q &= Q(local_site__name=local_site_name)
    else:
        q &= Q(local_site__isnull=True)

    return get_object_or_404(Repository, q)


@deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
def get_review_request_id(
    *,
    commit_message: str,
    server_url: str,
    commit_id: (str | None) = None,
    repository: (Repository | None) = None,
) -> int | None:
    """Return the review request ID matching the pushed commit.

    We first use a regex (that can be overridden in settings_local.py) to try
    to find a matching review request ID in the commit message. If no match is
    found with the regex, we then try to find a review request with a matching
    commit ID.

    We assume there is at most one review request associated with each commit.
    If a matching review request cannot be found, we return None.

    Version Changed:
        8.0:
        Made arguments keyword-only.

    Args:
        commit_message (str):
            The commit message from the hosting service.

        server_url (str):
            The URL to the Review Board server.

        commit_id (str):
            The ID of the commit.

        repository (reviewboard.scmtools.models.Repository):
            The repository for the commit.

    Returns:
        int:
        The ID of the review request that corresponds to the pushed commit. If
        no review request could be found, this will be ``None``.
    """
    regex = settings.HOSTINGSVCS_HOOK_REGEX % {
        'server_url': re.escape(server_url),
    }

    pattern = re.compile(regex, settings.HOSTINGSVCS_HOOK_REGEX_FLAGS)
    match = pattern.search(commit_message)

    if match:
        try:
            review_request_id = int(match.group('id'))
        except ValueError:
            logger.error('The review request ID must be an integer.')
            review_request_id = None
    elif commit_id:
        assert repository

        try:
            review_request = ReviewRequest.objects.get(
                commit_id=str(commit_id),
                repository=repository)
            review_request_id = review_request.display_id
        except ReviewRequest.DoesNotExist:
            review_request_id = None
    else:
        review_request_id = None

    return review_request_id


@deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
def close_review_request(
    *,
    review_request: ReviewRequest,
    review_request_id: int,
    description: str,
) -> None:
    """Close the specified review request as submitted.

    Version Changed:
        8.0:
        Made arguments keyword-only.

    Args:
        review_request (reviewboard.reviews.models.ReviewRequest):
            The review request to close.

        review_request_id (int):
            The ID of the review request.

        description (str):
            The close description to use.
    """
    if review_request.status == ReviewRequest.SUBMITTED:
        logger.warning('Review request #%s is already submitted.',
                       review_request_id)
        return

    # Closing as submitted will fail if the review request was never public. In
    # this case, publish first.
    if not review_request.public:
        review_request.publish(review_request.submitter, trivial=True,
                               validate_fields=False)

    review_request.close(ReviewRequest.SUBMITTED, description=description)
    logger.debug('Review request #%s is set to %s.',
                 review_request_id, review_request.status)


@deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
def close_all_review_requests(
    *,
    review_request_id_to_commits: Mapping[int | None, Sequence[str]],
    local_site_name: str | None,
    repository: Repository,
    hosting_service_id: str,
) -> None:
    """Close each review request in the given dictionary as submitted.

    The provided dictionary should map a review request ID (int) to commits
    associated with that review request ID (list of strings). Commits that are
    not associated with any review requests have the key None.

    Version Changed:
        8.0:
        Made arguments keyword-only.

    Args:
        review_request_id_to_commits (dict):
            A mapping from review request ID to a list of

        local_site_name (str):
            The Local Site name, if available.

        repository (reviewboard.scmtools.models.Repository):
            The repository for all review requests.

        hosting_service_id (str):
            The name of the hosting service.
    """
    if local_site_name:
        try:
            local_site = LocalSite.objects.get(name=local_site_name)
        except LocalSite.DoesNotExist:
            logger.error('close_all_review_requests: Local Site %s does '
                         'not exist.',
                         local_site_name)
            return
    else:
        local_site = None

    # Some of the entries we get may have 'None' keys, so filter them out.
    review_request_ids = [
        review_request_id
        for review_request_id in review_request_id_to_commits
        if review_request_id is not None
    ]

    if not review_request_ids:
        return

    # Look up all review requests that match the given repository, hosting
    # service ID, and Local Site.
    q = (Q(repository=repository) &
         Q(repository__hosting_account__service_name=hosting_service_id))

    if local_site:
        q &= Q(local_id__in=review_request_ids) & Q(local_site=local_site)
    else:
        q &= Q(pk__in=review_request_ids)

    review_requests = list(ReviewRequest.objects.filter(q))

    # Check if there are any listed that we couldn't find, and log them.
    if len(review_request_ids) != len(review_requests):
        id_to_review_request = {
            review_request.display_id: review_request
            for review_request in review_requests
        }

        for review_request_id in review_request_ids:
            if review_request_id not in id_to_review_request:
                logger.error('close_all_review_requests: Review request #%s '
                             'does not exist.',
                             review_request_id)

    # Close any review requests we did find.
    for review_request in review_requests:
        review_request_id = review_request.display_id

        commits_info = ', '.join(
            review_request_id_to_commits[review_request_id])

        close_review_request(
            review_request=review_request,
            review_request_id=review_request_id,
            description=f'Pushed to {commits_info}',
        )



#!/usr/bin/env python
#
# rbssh.py -- A custom SSH client for use in Review Board.
#
# This is used as an ssh replacement that can be used across platforms with
# a custom .ssh directory. OpenSSH doesn't respect $HOME, instead reading
# /etc/passwd directly, which causes problems for us. Using rbssh, we can
# work around this.
#
#
# Copyright (c) 2010-2011  Beanbag, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import errno
import getpass
import logging
import os
import select
import sys
import warnings
from optparse import OptionParser


# We don't want any warnings to end up impacting output.
warnings.simplefilter('ignore')

logger = logging.getLogger('rbssh')


if str('RBSITE_PYTHONPATH') in os.environ:
    for path in reversed(os.environ[str('RBSITE_PYTHONPATH')].split(str(':'))):
        sys.path.insert(1, path)

os.environ[str('DJANGO_SETTINGS_MODULE')] = \
    str('reviewboard.cmdline.conf.rbssh.settings')

import django
import paramiko

import reviewboard
from reviewboard import get_version_string


DEBUG = os.getenv('RBSSH_DEBUG') or os.getenv('DEBUG_RBSSH')
DEBUG_LOGDIR = os.getenv('RBSSH_LOG_DIR')
STORAGE_BACKEND = os.getenv('RBSSH_STORAGE_BACKEND')

SSH_PORT = 22

# This is the maximum size we'll read from the buffer. If there's less
# data available than this, it won't block, it'll just return the
# available data.
BUFFER_SIZE = 16384


options = None


if DEBUG:
    debug = logger.debug
else:
    debug = lambda *args, **kwargs: None


class PlatformHandler(object):
    """A generic base class for wrapping platform-specific operations.

    This should be subclassed for each major platform.
    """

    def __init__(self, channel):
        """Initialize the handler.

        Args:
            channel (paramiko.channel.Channel):
                The channel to process.
        """
        self.channel = channel
        self.stdin_fd = sys.stdin.fileno()
        self.stdout_fd = sys.stdout.fileno()
        self.stderr_fd = sys.stderr.fileno()

    def shell(self):
        """Open a shell."""
        raise NotImplementedError

    def transfer(self):
        """Transfer data over the channel."""
        raise NotImplementedError

    def write_input(self, data):
        """Write input to a channel.

        This will write to the channel, ensuring the full contents of the data
        have been written before this returns.

        This is used for stdin.

        Version Added:
            4.0.11

        Args:
            data (bytes):
                The data to write.
        """
        channel = self.channel
        written = 0

        while written < len(data):
            to_write = data[written:]

            debug('Writing stdin data=%r', to_write)

            try:
                written += channel.send(to_write)
            except OSError as e:
                if e.errno == errno.EPIPE:
                    # The pipe is closed. We're done.
                    debug('... pipe closed. Giving up')
                    break
                else:
                    debug('... blocked (errno=%s). Trying again.', e)

    def write_output(self, fd, data):
        """Write output to a file descriptor.

        This will write to the file descriptor, handling blocking I/O errors
        and ensuring the full contents of the data have been written before
        this returns.

        This is used for stdout and stderr.

        Args:
            fd (int):
                The file descriptor.

            data (bytes):
                The data to write.
        """
        written = 0

        while written < len(data):
            to_write = data[written:]

            debug('Writing fd=%r, output=%r', fd, to_write)

            try:
                written += os.write(fd, to_write)
            except OSError as e:
                if e.errno == errno.EPIPE:
                    # The pipe is closed. We're done.
                    debug('... pipe closed. Giving up')
                    break
                else:
                    debug('... blocked (errno=%s). Trying again.', e)

    def process_channel(self, channel):
        """Process the given channel.

        This will retrieve any data from the output and error streams and
        output it to stdout and stderr.

        Processing will finish when no new data can be read, no new data is
        available to read, and an exit status is available.

        Args:
            channel (paramiko.channel.Channel):
                The channel to process.

        Returns:
            bool:
            ``True`` if the channel should continue to be processed.
            ``False`` if processing has completed.
        """
        debug('== process_channel ==')

        has_data = False

        if channel.recv_ready():
            data = channel.recv(BUFFER_SIZE)

            if data:
                debug('got stdout=%r' % data)
                has_data = True

                self.write_output(self.stdout_fd, data)

        if channel.recv_stderr_ready():
            data = channel.recv_stderr(BUFFER_SIZE)

            if data:
                debug('got stderr=%r' % data)
                has_data = True

                self.write_output(self.stderr_fd, data)

        if (not has_data and
            channel.exit_status_ready() and
            not channel.recv_ready() and
            not channel.recv_stderr_ready()):
            # There should truly be nothing left to process. Everything has
            # indicated that the communication is done.
            debug('no data to read; exit ready; channel closed.')
            return False

        return True

    def process_stdin(self):
        """Read data from stdin and send it over the channel.

        Version Changed:
            4.0.11:
            Removed the ``channel`` argument.

        Returns:
            bool:
            ``True`` if the channel should continue to be processed.
            ``False`` if processing has completed.
        """
        debug('== process_stdin ==')

        try:
            buf = os.read(self.stdin_fd, BUFFER_SIZE)
        except OSError:
            buf = None

        if not buf:
            debug('stdin empty')
            return False

        self.write_input(buf)

        return True


class PosixHandler(PlatformHandler):
    """A platform handler for POSIX-type platforms."""

    def shell(self):
        """Open a shell."""
        import termios
        import tty

        stdin_fd = self.stdin_fd
        oldtty = termios.tcgetattr(sys.stdin)

        try:
            tty.setraw(stdin_fd)
            tty.setcbreak(stdin_fd)

            self.handle_communications()
        finally:
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, oldtty)

    def transfer(self):
        """Transfer data over the channel."""
        # Set all streams to be non-blocking. We'll manage the reading/writing
        # accordingly, handling any blocking I/O errors. This will ensure
        # that we don't buffer too long and risk any communication issues.
        for stream in (sys.stdin, sys.stdout, sys.stderr):
            os.set_blocking(stream.fileno(), False)

        self.handle_communications()

    def handle_communications(self):
        """Handle any pending data over the channel or stdin."""
        channel = self.channel
        stdin = sys.stdin
        to_read = [channel, stdin]

        while to_read:
            rl, wl, el = select.select(to_read, [], [])

            if channel in rl:
                if not self.process_channel(channel):
                    # Output streams are closed, and the handler told us
                    # that the channel is ready to exit. Break the loop.
                    break

            if stdin in rl:
                if not self.process_stdin():
                    channel.shutdown_write()
                    to_read.remove(stdin)


class WindowsHandler(PlatformHandler):
    """A platform handler for Microsoft Windows platforms."""

    def shell(self):
        """Open a shell."""
        self.handle_communications()

    def transfer(self):
        """Transfer data over the channel."""
        self.handle_communications()

    def handle_communications(self):
        """Handle any pending data over the channel or stdin."""
        import threading

        debug('begin_windows_transfer')

        self.channel.setblocking(0)

        def writeall(channel):
            while self.process_channel(channel):
                pass

            debug('Shutting down reading')
            channel.shutdown_read()

        writer = threading.Thread(target=writeall, args=(self.channel,))
        writer.start()

        try:
            while self.process_stdin():
                pass
        except EOFError:
            pass

        debug('Shutting down writing')
        self.channel.shutdown_write()


def print_version(option, opt, value, parser):
    """Print the current version and exit."""
    parser.print_version()
    sys.exit(0)


def parse_options(args):
    """Parse the given arguments into the global ``options`` dictionary."""
    global options

    hostname = None

    # NOTE: Update to use RBProgVersionAction when this is ported to argparse.
    parser = OptionParser(
        usage='%prog [options] [user@]hostname [command]',
        version=(
            '%%prog %s\n'
            'Python %s\n'
            'Installed to %s'
            % (get_version_string(),
               sys.version.splitlines()[0],
               os.path.dirname(reviewboard.__file__))
        ))
    parser.disable_interspersed_args()
    parser.add_option('-l',
                      dest='username', metavar='USERNAME', default=None,
                      help='the user to log in as on the remote machine')
    parser.add_option('-p', '--port',
                      type='int', dest='port', metavar='PORT', default=None,
                      help='the port to connect to')
    parser.add_option('-q', '--quiet',
                      action='store_true', dest='quiet', default=False,
                      help='suppress any unnecessary output')
    parser.add_option('-s',
                      dest='subsystem', metavar='SUBSYSTEM', default=None,
                      nargs=2,
                      help='the subsystem to use (ssh or sftp)')
    parser.add_option('-V',
                      action='callback', callback=print_version,
                      help='display the version information and exit')
    parser.add_option('--rb-disallow-agent',
                      action='store_false', dest='allow_agent',
                      default=os.getenv('RBSSH_ALLOW_AGENT') != '0',
                      help='disable using the SSH agent for authentication')
    parser.add_option('--rb-local-site',
                      dest='local_site_name', metavar='NAME',
                      default=os.getenv('RB_LOCAL_SITE'),
                      help='the local site name containing the SSH keys to '
                           'use')

    (options, args) = parser.parse_args(args)

    if options.subsystem:
        if len(options.subsystem) != 2:
            parser.error('-s requires a hostname and a valid subsystem')
        elif options.subsystem[1] not in ('sftp', 'ssh'):
            parser.error('Invalid subsystem %s' % options.subsystem[1])

        hostname, options.subsystem = options.subsystem

    if len(args) == 0 and not hostname:
        parser.print_help()
        sys.exit(1)

    if not hostname:
        hostname = args[0]
        args = args[1:]

    if options.port:
        port = options.port
    else:
        port = SSH_PORT

    return hostname, port, args


def main():
    """Run the application."""
    # Perform the bare minimum to initialize the Django/Review Board
    # environment. We're not calling Review Board's initialize() because
    # we want to completely minimize what we import and set up.
    if hasattr(django, 'setup'):
        django.setup()

    if DEBUG:
        pid = os.getpid()
        log_filename = 'rbssh-%s.log' % pid

        if DEBUG_LOGDIR:
            log_path = os.path.join(DEBUG_LOGDIR, log_filename)
        else:
            log_path = log_filename

        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s %(name)-18s %(levelname)-8s '
                                   '%(message)s',
                            datefmt='%m-%d %H:%M',
                            filename=log_path,
                            filemode='w')

        debug('rbssh %s', get_version_string())
        debug('Installed to %s',
              os.path.dirname(reviewboard.__file__))
        debug('Python %s' % sys.version.splitlines()[0])
        debug('PID %s', pid)
        debug('Command: %s', sys.argv)

    from reviewboard.scmtools.core import SCMTool
    from reviewboard.ssh.client import SSHClient

    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    ch.setFormatter(logging.Formatter('%(message)s'))
    ch.addFilter(logging.Filter('root'))
    logging.getLogger('').addHandler(ch)

    path, port, command = parse_options(sys.argv[1:])

    if '://' not in path:
        path = 'ssh://' + path

    username, hostname = SCMTool.get_auth_from_uri(path, options.username)

    if username is None:
        username = getpass.getuser()

    client = SSHClient(namespace=options.local_site_name,
                       storage_backend=STORAGE_BACKEND)
    client.set_missing_host_key_policy(paramiko.WarningPolicy())

    if command:
        purpose = command
    else:
        purpose = 'interactive shell'

    debug('SSH backend = %s', type(client.storage))
    debug('Preparing to connect to %s@%s for %s',
          username, hostname, purpose)

    attempts = 0
    password = None

    key = client.get_user_key()

    while True:
        try:
            client.connect(hostname, port, username=username,
                           password=password, pkey=key,
                           allow_agent=options.allow_agent)
            break
        except paramiko.AuthenticationException:
            if attempts == 3 or not sys.stdin.isatty():
                logger.error('Too many authentication failures for %s',
                             username)
                sys.stderr.write('Too many authentication failures for %s\n'
                                 % username)
                sys.exit(1)

            attempts += 1
            password = getpass.getpass("%s@%s's password: " %
                                       (username, hostname))
        except paramiko.SSHException as e:
            logger.error('Error connecting to server: %s', e)
            sys.stderr.write('Error connecting to server: %s\n' % e)
            sys.exit(1)
        except Exception as e:
            logger.exception('Unknown exception during connect: %s (%s)',
                             e, type(e))
            sys.stderr.write('Unknown exception during connect: %s (%s)\n'
                             % (e, type(e)))
            sys.exit(1)

    transport = client.get_transport()
    channel = transport.open_session()

    if sys.platform in ('cygwin', 'win32'):
        debug('Using WindowsHandler')
        handler = WindowsHandler(channel)
    else:
        debug('Using PosixHandler')
        handler = PosixHandler(channel)

    if options.subsystem == 'sftp':
        debug('Invoking sftp subsystem')
        channel.invoke_subsystem('sftp')
        handler.transfer()
    elif command:
        debug('Sending command %s', command)
        channel.exec_command(' '.join(command))
        handler.transfer()
    else:
        debug('Opening shell')
        channel.get_pty()
        channel.invoke_shell()
        handler.shell()

    debug('Execution has finished')
    status = channel.recv_exit_status()
    client.close()

    debug('Exit code = %s', status)
    debug('Shutting down')

    # Flush everything.
    try:
        logging.shutdown()
    except Exception:
        pass

    return status


if __name__ == '__main__':
    sys.exit(main())



"""Static bundles for Review Board."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping

    from djblets.pipeline import StaticBundle


PIPELINE_JAVASCRIPT: Mapping[str, StaticBundle] = {
    '3rdparty-admin': {
        'source_filenames': (
            'lib/js/3rdparty-admin/index.ts',
        ),
        'output_filename': 'lib/js/3rdparty-admin.min.js',
    },
    '3rdparty-base': {
        'source_filenames': (
            'lib/js/3rdparty-base/index.ts',
        ),
        'output_filename': 'lib/js/3rdparty-base.min.js',
    },
    '3rdparty': {
        'source_filenames': (
            'lib/js/3rdparty/index.ts',
        ),
        'output_filename': 'lib/js/3rdparty.min.js',
    },
    '3rdparty-jsonlint': {
        'source_filenames': (
            'lib/js/jsonlint/index.ts',
        ),
        'output_filename': 'lib/js/3rdparty-jsonlint.min.js',
    },
    'js-test-libs': {
        'source_filenames': (
            'lib/js/js-test-libs/index.ts',
        ),
        'output_filename': 'rb/js/js-test-libs.min.js',
    },
    'js-tests': {
        'source_filenames': (
            'rb/js/tests/index.ts',
            'rb/js/admin/tests/relatedGroupSelectorViewTests.es6.js',
            'rb/js/admin/tests/relatedRepoSelectorViewTests.es6.js',
            'rb/js/admin/tests/relatedUserSelectorViewTests.es6.js',
            'rb/js/admin/models/tests/dashboardPageModelTests.es6.js',
            'rb/js/admin/models/tests/inlineFormGroupModelTests.es6.js',
            'rb/js/admin/models/tests/newsWidgetModelTests.es6.js',
            'rb/js/admin/views/tests/dashboardPageViewTests.es6.js',
            'rb/js/admin/views/tests/newsWidgetViewTests.es6.js',
            'rb/js/admin/views/tests/inlineFormGroupViewTests.es6.js',
            'rb/js/admin/views/tests/inlineFormViewTests.es6.js',
            'rb/js/collections/tests/filteredCollectionTests.es6.js',
            'rb/js/configForms/models/tests/resourceListItemModelTests.es6.js',
            'rb/js/diffviewer/models/tests/diffCommitModelTests.es6.js',
            'rb/js/diffviewer/models/tests/diffRevisionModelTests.es6.js',
            'rb/js/diffviewer/models/tests/paginationModelTests.es6.js',
            'rb/js/diffviewer/views/tests/diffCommitListViewTests.es6.js',
            'rb/js/models/tests/uploadDiffModelTests.es6.js',
            'rb/js/newReviewRequest/views/tests/branchesViewTests.es6.js',
            'rb/js/newReviewRequest/views/tests/postCommitViewTests.es6.js',
            'rb/js/newReviewRequest/views/tests/repositorySelectionViewTests.es6.js',
            'rb/js/pages/models/tests/pageManagerModelTests.es6.js',
            'rb/js/resources/collections/tests/repositoryBranchesCollectionTests.es6.js',
            'rb/js/resources/collections/tests/repositoryCommitsCollectionTests.es6.js',
            'rb/js/resources/models/tests/baseCommentReplyModelTests.es6.js',
            'rb/js/resources/models/tests/diffCommentModelTests.es6.js',
            'rb/js/resources/models/tests/draftReviewRequestModelTests.es6.js',
            'rb/js/resources/models/tests/fileAttachmentCommentModelTests.es6.js',
            'rb/js/resources/models/tests/fileDiffModelTests.es6.js',
            'rb/js/resources/models/tests/generalCommentModelTests.es6.js',
            'rb/js/resources/models/tests/screenshotModelTests.es6.js',
            'rb/js/resources/models/tests/screenshotCommentModelTests.es6.js',
            'rb/js/resources/models/tests/repositoryBranchModelTests.es6.js',
            'rb/js/resources/models/tests/repositoryCommitModelTests.es6.js',
            'rb/js/resources/models/tests/userFileAttachmentModelTests.es6.js',
            'rb/js/resources/models/tests/validateDiffModelTests.es6.js',
            'rb/js/reviewRequestPage/models/tests/changeEntryModelTests.es6.js',
            'rb/js/reviewRequestPage/models/tests/entryModelTests.es6.js',
            'rb/js/reviewRequestPage/models/tests/reviewEntryModelTests.es6.js',
            'rb/js/reviewRequestPage/models/tests/reviewReplyEditorModelTests.es6.js',
            'rb/js/reviewRequestPage/models/tests/reviewRequestPageModelTests.es6.js',
            'rb/js/reviewRequestPage/models/tests/statusUpdatesEntryModelTests.es6.js',
            'rb/js/reviewRequestPage/views/tests/baseStatusUpdatesEntryViewTests.es6.js',
            'rb/js/reviewRequestPage/views/tests/issueSummaryTableViewTests.es6.js',
            'rb/js/reviewRequestPage/views/tests/reviewEntryViewTests.es6.js',
            'rb/js/reviewRequestPage/views/tests/reviewReplyDraftBannerViewTests.es6.js',
            'rb/js/reviewRequestPage/views/tests/reviewReplyEditorViewTests.es6.js',
            'rb/js/reviewRequestPage/views/tests/reviewRequestPageViewTests.es6.js',
            'rb/js/reviewRequestPage/views/tests/reviewViewTests.es6.js',
            'rb/js/ui/views/tests/dialogViewTests.es6.js',
            'rb/js/ui/views/tests/drawerViewTests.es6.js',
            'rb/js/ui/views/tests/formViewTests.es6.js',
            'rb/js/ui/views/tests/infoboxManagerViewTests.es6.js',
            'rb/js/ui/views/tests/menuButtonViewTests.es6.js',
            'rb/js/ui/views/tests/menuViewTests.es6.js',
            'rb/js/ui/views/tests/notificationManagerTests.es6.js',
            'rb/js/ui/views/tests/scrollManagerViewTests.es6.js',
            'rb/js/utils/tests/apiUtilsTests.es6.js',
            'rb/js/utils/tests/dataUtilsTests.es6.js',
            'rb/js/utils/tests/keyBindingUtilsTests.es6.js',
            'rb/js/utils/tests/linkifyUtilsTests.es6.js',
            'rb/js/utils/tests/urlUtilsTests.es6.js',
            'rb/js/views/tests/clientLoginViewTests.es6.js',
            'rb/js/views/tests/collectionViewTests.es6.js',
            'rb/js/views/tests/diffFragmentQueueViewTests.es6.js',
            'rb/js/views/tests/diffFragmentViewTests.es6.js',
            'rb/js/views/tests/draftReviewBannerViewTests.es6.js',
            'rb/js/views/tests/screenshotThumbnailViewTests.es6.js',
            'rb/js/views/tests/uploadAttachmentViewTests.es6.js',
        ),
        'output_filename': 'rb/js/js-tests.min.js',
    },
    'common': {
        'source_filenames': (
            # Include the common ESM bundle and the extensions bundle. This
            # will ensure they're both exposed as "RB", but allow
            # `reviewboard/extensions` to be a separate ESM bundle that
            # extension code can target.
            'rb/js/common/index.ts',
            'rb/js/extensions/index.ts',

            # Legacy JavaScript
            'rb/js/utils/underscoreUtils.es6.js',
            'rb/js/init.es6.js',
            'rb/js/utils/apiErrors.es6.js',
            'rb/js/utils/apiUtils.es6.js',
            'rb/js/utils/dataUtils.es6.js',
            'rb/js/utils/linkifyUtils.es6.js',
            'rb/js/utils/mathUtils.es6.js',
            'rb/js/utils/keyBindingUtils.es6.js',
            'rb/js/utils/urlUtils.es6.js',
            'rb/js/collections/filteredCollection.es6.js',
            'rb/js/pages/models/pageManagerModel.es6.js',
            'rb/js/resources/models/apiTokenModel.es6.js',
            'rb/js/resources/models/repositoryBranchModel.es6.js',
            'rb/js/resources/models/repositoryCommitModel.es6.js',
            'rb/js/resources/models/draftReviewRequestModel.es6.js',
            'rb/js/resources/models/baseCommentReplyModel.es6.js',
            'rb/js/resources/models/diffCommentModel.es6.js',
            'rb/js/resources/models/diffCommentReplyModel.es6.js',
            'rb/js/resources/models/diffModel.es6.js',
            'rb/js/resources/models/fileAttachmentCommentModel.es6.js',
            'rb/js/resources/models/fileAttachmentCommentReplyModel.es6.js',
            'rb/js/resources/models/generalCommentModel.es6.js',
            'rb/js/resources/models/generalCommentReplyModel.es6.js',
            'rb/js/resources/models/fileDiffModel.es6.js',
            'rb/js/resources/models/screenshotModel.es6.js',
            'rb/js/resources/models/screenshotCommentModel.es6.js',
            'rb/js/resources/models/screenshotCommentReplyModel.es6.js',
            'rb/js/resources/models/userFileAttachmentModel.es6.js',
            'rb/js/resources/models/validateDiffModel.es6.js',
            'rb/js/resources/collections/repositoryBranchesCollection.es6.js',
            'rb/js/resources/collections/repositoryCommitsCollection.es6.js',
            'rb/js/ui/views/dialogView.es6.js',
            'rb/js/ui/views/formView.es6.js',
            'rb/js/ui/views/baseInfoboxView.es6.js',
            'rb/js/ui/views/infoboxManagerView.es6.js',
            'rb/js/ui/views/bugInfoboxView.es6.js',
            'rb/js/ui/views/drawerView.es6.js',
            'rb/js/ui/views/notificationManager.es6.js',
            'rb/js/ui/views/reviewRequestInfoboxView.es6.js',
            'rb/js/ui/views/scrollManagerView.es6.js',
            'rb/js/ui/views/userInfoboxView.es6.js',
            'rb/js/models/starManagerModel.es6.js',
            'rb/js/views/clientLoginView.es6.js',
            'rb/js/views/headerView.es6.js',
            'rb/js/views/collectionView.es6.js',
            'rb/js/views/starManagerView.es6.js',
        ),
        'output_filename': 'rb/js/base.min.js',
    },
    'ui': {
        'source_filenames': (
            'rb/js/ui/index.ts',
        ),
        'output_filename': 'rb/js/ui.min.js',
    },
    'account-page': {
        'source_filenames': (
            'rb/js/accountPrefsPage/index.ts',

            # Legacy JavaScript
            'rb/js/accountPrefsPage/views/apiTokensView.es6.js',
            'rb/js/accountPrefsPage/views/joinedGroupsView.es6.js',
            'rb/js/accountPrefsPage/views/oauthApplicationsView.es6.js',
            'rb/js/accountPrefsPage/views/oauthTokensView.es6.js',
        ),
        'output_filename': 'rb/js/account-page.min.js',
    },
    'config-forms': {
        'source_filenames': (
            'rb/js/configForms/index.ts',

            # Legacy JavaScript
            'rb/js/configForms/models/resourceListItemModel.es6.js',
        ),
        'output_filename': 'rb/js/config-forms.min.js',
    },
    'datagrid-pages': {
        'source_filenames': (
            'rb/js/datagrids/index.ts',

            # Legacy JavaScript
            'rb/js/pages/models/datagridPageModel.es6.js',
            'rb/js/pages/models/dashboardModel.es6.js',
            'rb/js/pages/views/datagridPageView.es6.js',
            'rb/js/pages/views/dashboardView.es6.js',
        ),
        'output_filename': 'rb/js/dashboard.min.js',
    },
    'reviews': {
        'source_filenames': (
            'rb/js/reviews/index.ts',

            # Legacy JavaScript
            #
            # Note: These are roughly in dependency order.
            'rb/js/models/uploadDiffModel.es6.js',
            'rb/js/utils/textUtils.es6.js',
            'rb/js/views/diffFragmentQueueView.es6.js',
            'rb/js/views/diffFragmentView.es6.js',
            'rb/js/views/draftReviewBannerView.es6.js',
            'rb/js/views/uploadAttachmentView.es6.js',
            'rb/js/views/revisionSelectorView.es6.js',
            'rb/js/views/fileAttachmentRevisionLabelView.es6.js',
            'rb/js/views/fileAttachmentRevisionSelectorView.es6.js',
            'rb/js/views/screenshotThumbnailView.es6.js',
            'rb/js/views/uploadDiffView.es6.js',
            'rb/js/views/updateDiffView.es6.js',
            'rb/js/diffviewer/models/commitHistoryDiffEntry.es6.js',
            'rb/js/diffviewer/models/diffCommitListModel.es6.js',
            'rb/js/diffviewer/models/diffCommitModel.es6.js',
            'rb/js/diffviewer/models/diffRevisionModel.es6.js',
            'rb/js/diffviewer/models/paginationModel.es6.js',
            'rb/js/diffviewer/collections/commitHistoryDiffEntryCollection.es6.js',
            'rb/js/diffviewer/collections/diffCommitCollection.es6.js',
            'rb/js/diffviewer/views/chunkHighlighterView.es6.js',
            'rb/js/diffviewer/views/diffCommitListView.es6.js',
            'rb/js/diffviewer/views/diffRevisionLabelView.es6.js',
            'rb/js/diffviewer/views/diffRevisionSelectorView.es6.js',
            'rb/js/diffviewer/views/paginationView.es6.js',
        ),
        'output_filename': 'rb/js/reviews.min.js',
    },
    'review-request-page': {
        'source_filenames': (
            'rb/js/reviewRequestPage/index.ts',

            # Legacy JavaScript
            'rb/js/reviewRequestPage/models/entryModel.es6.js',
            'rb/js/reviewRequestPage/models/reviewEntryModel.es6.js',
            'rb/js/reviewRequestPage/models/reviewReplyEditorModel.es6.js',
            'rb/js/reviewRequestPage/models/reviewRequestPageModel.es6.js',
            'rb/js/reviewRequestPage/models/statusUpdatesEntryModel.es6.js',
            'rb/js/reviewRequestPage/models/changeEntryModel.es6.js',
            'rb/js/reviewRequestPage/views/entryView.es6.js',
            'rb/js/reviewRequestPage/views/baseStatusUpdatesEntryView.es6.js',
            'rb/js/reviewRequestPage/views/changeEntryView.es6.js',
            'rb/js/reviewRequestPage/views/initialStatusUpdatesEntryView.es6.js',
            'rb/js/reviewRequestPage/views/issueSummaryTableView.es6.js',
            'rb/js/reviewRequestPage/views/reviewEntryView.es6.js',
            'rb/js/reviewRequestPage/views/reviewReplyEditorView.es6.js',
            'rb/js/reviewRequestPage/views/reviewRequestPageView.es6.js',
            'rb/js/reviewRequestPage/views/reviewView.es6.js',
        ),
        'output_filename': 'rb/js/review-request-page.min.js',
    },
    'newReviewRequest': {
        'source_filenames': (
            'rb/js/newReviewRequest/index.ts',

            # Legacy JavaScript
            #
            # Note: These are roughly in dependency order.
            'rb/js/models/uploadDiffModel.es6.js',
            'rb/js/newReviewRequest/models/postCommitModel.es6.js',
            'rb/js/newReviewRequest/models/newReviewRequestModel.es6.js',
            'rb/js/newReviewRequest/collections/repositoryCollection.es6.js',
            'rb/js/views/uploadDiffView.es6.js',
            'rb/js/newReviewRequest/views/branchView.es6.js',
            'rb/js/newReviewRequest/views/branchesView.es6.js',
            'rb/js/newReviewRequest/views/commitView.es6.js',
            'rb/js/newReviewRequest/views/commitsView.es6.js',
            'rb/js/newReviewRequest/views/repositoryView.es6.js',
            'rb/js/newReviewRequest/views/repositorySelectionView.es6.js',
            'rb/js/newReviewRequest/views/postCommitView.es6.js',
            'rb/js/newReviewRequest/views/preCommitView.es6.js',
            'rb/js/newReviewRequest/views/newReviewRequestView.es6.js',
        ),
        'output_filename': 'rb/js/newReviewRequest.min.js',
    },
    'oauth-edit': {
        'source_filenames': (
            # Legacy JavaScript
            'rb/js/accountPrefsPage/views/oauthClientSecretView.es6.js',
        ),
        'output_filename': 'rb/js/oauth.min.js',
    },
    'admin': {
        'source_filenames': (
            'rb/js/admin/index.ts',

            # Legacy JavaScript
            'rb/js/admin/models/changeListPageModel.es6.js',
            'rb/js/admin/models/dashboardPageModel.es6.js',
            'rb/js/admin/models/inlineFormGroupModel.es6.js',
            'rb/js/admin/models/inlineFormModel.es6.js',
            'rb/js/admin/models/widgetModel.es6.js',
            'rb/js/admin/models/newsWidgetModel.es6.js',
            'rb/js/admin/models/serverActivityWidgetModel.es6.js',
            'rb/js/admin/views/changeFormPageView.es6.js',
            'rb/js/admin/views/changeListPageView.es6.js',
            'rb/js/admin/views/dashboardPageView.es6.js',
            'rb/js/admin/views/inlineFormGroupView.es6.js',
            'rb/js/admin/views/inlineFormView.es6.js',
            'rb/js/admin/views/widgetView.es6.js',
            'rb/js/admin/views/newsWidgetView.es6.js',
            'rb/js/admin/views/serverActivityWidgetView.es6.js',
            'rb/js/admin/views/userActivityWidgetView.es6.js',
        ),
        'output_filename': 'rb/js/admin.min.js',
    },
    'repositoryform': {
        'source_filenames': (
            # Legacy JavaScript
            'rb/js/admin/repositoryform.es6.js',
        ),
        'output_filename': 'rb/js/repositoryform.min.js',
    },
    'webhooks-form': {
        'source_filenames': (
            # Legacy JavaScript
            'rb/js/admin/views/webhookFormView.es6.js',
        ),
        'output_filename': 'rb/js/webhooks-form.min.js',
    },
    'widgets': {
        'source_filenames': (
            # Legacy JavaScript
            'rb/js/admin/views/relatedUserSelectorView.es6.js',
            'rb/js/admin/views/relatedRepoSelectorView.es6.js',
            'rb/js/admin/views/relatedGroupSelectorView.es6.js',
        ),
        'output_filename': 'rb/js/widgets.min.js',
    },
}


PIPELINE_STYLESHEETS: Mapping[str, StaticBundle] = {
    'common': {
        'source_filenames': (
            'lib/css/3rdparty.less',
            'lib/css/fontawesome.less',
            'rb/css/bundles/common.less',
        ),
        'output_filename': 'rb/css/common.min.css',
    },
    'js-tests': {
        'source_filenames': (
            'rb/css/pages/js-tests.less',
        ),
        'output_filename': 'rb/css/js-tests.min.css',
    },
    'account-page': {
        'source_filenames': (
            'rb/css/pages/my-account.less',
        ),
        'output_filename': 'rb/css/account-page.min.css',
    },
    'reviews': {
        'source_filenames': (
            'rb/css/bundles/reviews.less',
        ),
        'output_filename': 'rb/css/reviews.min.css',
    },
    'newReviewRequest': {
        'source_filenames': (
            'rb/css/pages/newReviewRequest.less',
        ),
        'output_filename': 'rb/css/newReviewRequest.min.css',
    },
    'oauth': {
        'source_filenames': (
            'rb/css/pages/oauth.less',
        ),
        'output_filename': 'rb/css/oauth.min.css',
    },
    'admin': {
        'source_filenames': (
            'rb/css/bundles/admin.less',
        ),
        'output_filename': 'rb/css/admin.min.css',
    },
}



"""Unit tests for reviewboard.site.urlresolvers."""

from django.http import HttpRequest

from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing.testcase import TestCase


class LocalSiteReverseTests(TestCase):
    """Unit tests for local_site_reverse."""

    def test_with_no_local_site(self):
        """Testing local_site_reverse with no local site"""
        request = HttpRequest()

        self.assertEqual(local_site_reverse('dashboard'),
                         '/dashboard/')
        self.assertEqual(local_site_reverse('dashboard', request=request),
                         '/dashboard/')
        self.assertEqual(local_site_reverse('user', args=['sample-user']),
                         '/users/sample-user/')
        self.assertEqual(
            local_site_reverse('user', kwargs={'username': 'sample-user'}),
            '/users/sample-user/')

    def test_with_local_site(self):
        """Testing local_site_reverse with a local site"""
        request = HttpRequest()
        request.GET['local_site_name'] = 'test'

        self.assertEqual(local_site_reverse('dashboard', request=request),
                         '/dashboard/')
        self.assertEqual(local_site_reverse('user', args=['sample-user'],
                                            request=request),
                         '/users/sample-user/')
        self.assertEqual(
            local_site_reverse('user', kwargs={'username': 'sample-user'},
                               request=request),
            '/users/sample-user/')



"""Unit tests for reviewboard.datagrids.columns.ReviewSummaryColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import ReviewSummaryColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class ReviewSummaryColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.ReviewSummaryColumn.

    Version Added:
        8.0
    """

    column = ReviewSummaryColumn()

    def test_render_data(self) -> None:
        """Testing ReviewSummaryColumn.render_data"""
        review_request = self.create_review_request(
            summary='Test Summary',
            publish=True)
        review = self.create_review(review_request)

        value = self.column.render_data(self.stateful_column, review)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, 'Test Summary')

    def test_render_data_with_special_chars(self) -> None:
        """Testing ReviewSummaryColumn.render_data with special HTML chars"""
        review_request = self.create_review_request(
            summary='<script>xss</script>',
            publish=True)
        review = self.create_review(review_request)

        value = self.column.render_data(self.stateful_column, review)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '&lt;script&gt;xss&lt;/script&gt;')

    def test_render_data_with_empty_summary(self) -> None:
        """Testing ReviewSummaryColumn.render_data with empty summary"""
        review_request = self.create_review_request(
            summary='',
            publish=True)
        review = self.create_review(review_request)

        value = self.column.render_data(self.stateful_column, review)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_to_json(self) -> None:
        """Testing ReviewSummaryColumn.to_json"""
        review_request = self.create_review_request(
            summary='Test Summary',
            publish=True)
        review = self.create_review(review_request)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review),
            'Test Summary')

    def test_to_json_does_not_escape_html(self) -> None:
        """Testing ReviewSummaryColumn.to_json does not escape HTML"""
        review_request = self.create_review_request(
            summary='<script>xss</script>',
            publish=True)
        review = self.create_review(review_request)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review),
            '<script>xss</script>')

    def test_to_json_with_empty_summary(self) -> None:
        """Testing ReviewSummaryColumn.to_json with empty summary"""
        review_request = self.create_review_request(
            summary='',
            publish=True)
        review = self.create_review(review_request)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review),
            '')



import logging
import os
import re
import subprocess
from tempfile import mkstemp

from django.utils.translation import gettext_lazy as _
from djblets.util.filesystem import is_exe_in_path

from reviewboard.scmtools.core import (SCMTool, ChangeSet,
                                       HEAD, PRE_CREATION)
from reviewboard.scmtools.errors import (SCMError, FileNotFoundError,
                                         RepositoryNotFoundError)
from reviewboard.diffviewer.parser import DiffParser


logger = logging.getLogger(__name__)


class PlasticTool(SCMTool):
    scmtool_id = 'plastic'
    name = "Plastic SCM"
    diffs_use_absolute_paths = True
    supports_pending_changesets = True
    field_help_text = {
        'path': _('The Plastic repository spec in the form of '
                  '[repo]@[hostname]:[port].'),
    }
    dependencies = {
        'executables': ['cm'],
    }

    REP_RE = re.compile(br'^(?P<reponame>.*)@(?P<hostname>.*):(?P<port>\d+)$')
    CS_RE = re.compile(br'^(?P<csid>\d+) (?P<user>[^\s]+) (?P<revid>\d+) '
                       br'(?P<file>.*)$')
    REPOLIST_RE = re.compile(br'^\s*\d+\s*(?P<reponame>[^\s]+)\s*.*:.*$')
    UNKNOWN_REV = "rev:revid:-1"

    def __init__(self, repository):
        super(PlasticTool, self).__init__(repository)

        self.reponame, self.hostname, self.port = \
            self.parse_repository(repository.path)
        self.client = PlasticClient(repository.path, self.reponame,
                                    self.hostname, self.port)

    def get_changeset(self, changesetid, allow_empty=False):
        logger.debug('Plastic: get_changeset %s', changesetid)

        changesetdata = self.client.get_changeset(changesetid)
        logger.debug('Plastic: changesetdata %s', changesetdata)

        # Changeset data is in the form of multiple lines of:
        # <changesetid> <user> <revid> <file spec>
        #
        # We assume the user and comment will be the same for each item, so
        # read it out of the first.
        #

        changeset = ChangeSet()
        changeset.changenum = changesetid

        split = changesetdata.split(b'\n')
        m = self.CS_RE.match(split[0])
        revid = m.group("revid")
        changeset.username = m.group("user")
        changeset.summary = self.client.get_changeset_comment(changesetid,
                                                              revid)
        logger.debug('Plastic: changeset user %s summary %s',
                     changeset.username, changeset.summary)

        for line in split:
            if line:
                m = self.CS_RE.match(line)

                if not m:
                    logger.debug('Plastic: bad re %s failed to match %s',
                                 self.CS_RE, line)
                    raise SCMError("Error looking up changeset")

                if m.group('csid') != str(changesetid):
                    logger.debug('Plastic: csid %s != %s',
                                 m.group('csid'), changesetid)
                    raise SCMError('The server returned a changeset ID that '
                                   'was not requested')

                logger.debug('Plastic: adding file %s',
                             m.group('file'))
                changeset.files += m.group('file')

        return changeset

    def get_file(self, path, revision=HEAD, **kwargs):
        logger.debug('Plastic: get_file %s revision %s', path, revision)

        if revision == PRE_CREATION:
            return b''

        # Check for new files
        if revision == self.UNKNOWN_REV:
            return b''

        return self.client.get_file(path, revision)

    def file_exists(self, path, revision=HEAD, **kwargs):
        logger.debug('Plastic: file_exists %s revision %s', path, revision)

        if revision == PRE_CREATION:
            return True

        # Check for new files
        if revision == self.UNKNOWN_REV:
            return True

        try:
            return self.client.get_file(path, revision)
        except FileNotFoundError:
            return False

    def parse_diff_revision(self, filename, revision, *args, **kwargs):
        """Parse and return a filename and revision from a diff.

        Args:
            filename (bytes):
                The filename as represented in the diff.

            revision (bytes):
                The revision as represented in the diff.

            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            tuple:
            A tuple containing two items:

            1. The normalized filename as a byte string.
            2. The normalized revision as a byte string or a
               :py:class:`~reviewboard.scmtools.core.Revision`.
        """
        assert isinstance(filename, bytes), (
            'filename must be a byte string, not %s' % type(filename))
        assert isinstance(revision, bytes), (
            'revision must be a byte string, not %s' % type(revision))

        logger.debug('Plastic: parse_diff_revision file %s revision %s',
                     filename.decode('utf-8'),
                     revision.decode('utf-8'))

        if revision == b'PRE-CREATION':
            revision = PRE_CREATION

        return filename, revision

    def get_parser(self, data):
        return PlasticDiffParser(data)

    @classmethod
    def parse_repository(cls, path):
        m = cls.REP_RE.match(path)

        if m:
            repopath = m.group("reponame")
            hostname = m.group("hostname")
            port = m.group("port")

            return repopath, hostname, port
        else:
            raise RepositoryNotFoundError()

    @classmethod
    def check_repository(cls, path, **kwargs):
        """Perform checks on a repository to test its validity.

        This checks if a repository exists and can be connected to.

        A failed result is returned as an exception. The exception may contain
        extra information, such as a human-readable description of the problem.
        If the repository is valid and can be connected to, no exception will
        be thrown.

        Args:
            path (unicode):
                The repository path.

            **kwargs (dict, unused):
                Additional settings for the repository.

        Raises:
            reviewboard.scmtools.errors.RepositoryNotFoundError:
                The repository at the given path could not be found.

            reviewboard.scmtools.errors.SCMError:
                There was a general error communicating with Perforce.
        """
        m = cls.REP_RE.match(path)

        if not m:
            raise RepositoryNotFoundError()

        # Can't use 'cm checkconnection' here as it only checks the
        # pre-configured server

        server = "%s:%s" % (m.group("hostname"), m.group("port"))
        reponame = m.group("reponame")

        logger.debug('Plastic: Checking repository %s@%s',
                     reponame, server)

        repositories = PlasticClient.get_repositories(server)
        split = repositories.splitlines()

        for rep in split:
            m = cls.REPOLIST_RE.match(rep)
            if m and m.group("reponame") == reponame:
                break
        else:
            raise RepositoryNotFoundError()


class PlasticDiffParser(DiffParser):
    """
    This class is able to parse diffs created with the plastic client
    support in post-review.
    """

    # As the diff creation is based on the Perforce code, so this is based
    # on the PerforceDiffParser (specifically, the binary file markers)
    BINARY_RE = re.compile(br'^==== ([^\s]+) \(([^\)]+)\) ==([ACIMR])==$')

    def __init__(self, data):
        super(PlasticDiffParser, self).__init__(data)

    def parse_diff_header(self, linenum, info):
        m = self.BINARY_RE.match(self.lines[linenum])

        if m:
            info['origFile'] = m.group(1)
            info['origInfo'] = m.group(2)
            info['newFile'] = m.group(1)
            info['newInfo'] = ""
            linenum += 1

            if (linenum < len(self.lines) and
                (self.lines[linenum].startswith(b"Binary files ") or
                 self.lines[linenum].startswith(b"Files "))):
                info['binary'] = True
                linenum += 1

            # In this case, this *is* our diff header.  We don't want to
            # let the next line's real diff header be a part of this one,
            # so return now
            return linenum

        return super(PlasticDiffParser, self).parse_diff_header(linenum, info)


class PlasticClient(object):
    def __init__(self, repository, reponame, hostname, port):
        if not is_exe_in_path('cm'):
            # This is technically not the right kind of error, but it's the
            # pattern we use with all the other tools.
            raise ImportError

        self.reponame = reponame
        self.hostname = hostname
        self.port = port

    def get_file(self, path, revision):
        logger.debug('Plastic: get_file %s rev %s', path, revision)

        repo = "rep:%s@repserver:%s:%s" % (self.reponame, self.hostname,
                                           self.port)

        # Work around a plastic bug, where 'cm cat --file=blah' gets an
        # extra newline, but plain 'cm cat' doesn't
        fd, tmpfile = mkstemp()
        os.close(fd)

        p = subprocess.Popen(
            ['cm', 'cat', revision + '@' + repo, '--file=' + tmpfile],
            stderr=subprocess.PIPE, stdout=subprocess.PIPE,
            close_fds=(os.name != 'nt'))
        errmsg = str(p.stderr.read())
        failure = p.wait()

        if failure:
            if not errmsg:
                errmsg = p.stdout.read()

            raise SCMError(errmsg)

        with open(tmpfile, 'rb') as readtmp:
            contents = readtmp.read()
        os.unlink(tmpfile)

        return contents

    def get_changeset(self, changesetid):
        logger.debug('Plastic: get_changeset %s', changesetid)

        repo = "rep:%s@repserver:%s:%s" % (self.reponame, self.hostname,
                                           self.port)

        p = subprocess.Popen(['cm', 'find', 'revs', 'where',
                              'changeset=' + str(changesetid), 'on',
                              'repository', '\'' + repo + '\'',
                              '--format={changeset} {owner} {id} {item}',
                              '--nototal'],
                             stderr=subprocess.PIPE, stdout=subprocess.PIPE,
                             close_fds=(os.name != 'nt'))
        contents = p.stdout.read()
        errmsg = p.stderr.read()
        failure = p.wait()

        if failure:
            raise SCMError(errmsg)

        return contents

    def get_changeset_comment(self, changesetid, revid):
        logger.debug('Plastic: get_changeset_comment %s', changesetid)

        repo = "rep:%s@repserver:%s:%s" % (self.reponame, self.hostname,
                                           self.port)

        p = subprocess.Popen(['cm', 'find', 'changesets', 'where',
                              'changesetid=' + str(changesetid),
                              'on', 'repository', '\'' + repo + '\'',
                              '--format={comment}', '--nototal'],
                             stderr=subprocess.PIPE, stdout=subprocess.PIPE,
                             close_fds=(os.name != 'nt'))
        contents = p.stdout.read()
        errmsg = p.stderr.read()
        failure = p.wait()

        if failure:
            raise SCMError(errmsg)

        return contents

    @classmethod
    def get_repositories(cls, server):
        logger.debug('Plastic: get_repositories %s', server)

        p = subprocess.Popen(['cm', 'listrepositories', server],
                             stderr=subprocess.PIPE, stdout=subprocess.PIPE,
                             close_fds=(os.name != 'nt'))
        repositories = p.stdout.read()
        errmsg = p.stderr.read()
        failure = p.wait()

        if failure:
            if not errmsg and repositories.startswith('Error:'):
                error = repositories
            else:
                error = errmsg

            raise SCMError(error)

        return repositories



"""Unit tests for reviewboard.certs.cert.Certificate.

Version Added:
    6.0
"""

from __future__ import annotations

import ipaddress
import json
import os
import tempfile
from datetime import datetime, timedelta, timezone

import kgb
from cryptography.x509.oid import NameOID
from djblets.util.serializers import DjbletsJSONEncoder

from reviewboard.certs.cert import (CertPurpose,
                                    Certificate,
                                    CertificateFingerprints)
from reviewboard.certs.errors import (CertificateNotFoundError,
                                      CertificateStorageError,
                                      InvalidCertificateFormatError)
from reviewboard.certs.tests.testcases import (CertificateTestCase,
                                               TEST_CLIENT_CERT_PEM,
                                               TEST_CLIENT_KEY_PEM,
                                               TEST_SHA1,
                                               TEST_SHA256,
                                               TEST_TRUST_CERT_PEM)
from reviewboard.testing import online_only


class CertificateTests(kgb.SpyAgency, CertificateTestCase):
    """Unit tests for Certificate.

    Version Added:
        6.0
    """

    def test_create_from_files_with_trust(self) -> None:
        """Testing Certificate.create_from_files with purpose=trust"""
        cert_fd, cert_path = tempfile.mkstemp()
        os.write(cert_fd, TEST_TRUST_CERT_PEM)
        os.close(cert_fd)

        try:
            cert = Certificate.create_from_files(
                hostname='example.com',
                port=443,
                cert_path=cert_path)
        finally:
            os.unlink(cert_path)

        self.assertEqual(cert.cert_data, TEST_TRUST_CERT_PEM)
        self.assertIsNone(cert.key_data)

    def test_create_from_files_with_trust_and_key(self) -> None:
        """Testing Certificate.create_from_files with purpose=trust and
        key_path=
        """
        cert_fd, cert_path = tempfile.mkstemp()
        os.write(cert_fd, TEST_TRUST_CERT_PEM)
        os.close(cert_fd)

        key_fd, key_path = tempfile.mkstemp()
        os.write(key_fd, TEST_CLIENT_KEY_PEM)
        os.close(key_fd)

        message = (
            'key_path cannot be provided when passing '
            'purpose=CertPurpose.TRUST.'
        )

        try:
            with self.assertRaisesMessage(ValueError, message):
                Certificate.create_from_files(
                    hostname='example.com',
                    port=443,
                    purpose=CertPurpose.TRUST,
                    cert_path=cert_path,
                    key_path=key_path,
                )
        finally:
            os.unlink(cert_path)
            os.unlink(key_path)

    def test_create_from_files_with_client_and_key(self) -> None:
        """Testing Certificate.create_from_files with purpose=client and
        key_path=
        """
        cert_fd, cert_path = tempfile.mkstemp()
        os.write(cert_fd, TEST_CLIENT_CERT_PEM)
        os.close(cert_fd)

        key_fd, key_path = tempfile.mkstemp()
        os.write(key_fd, TEST_CLIENT_KEY_PEM)
        os.close(key_fd)

        try:
            cert = Certificate.create_from_files(
                hostname='example.com',
                port=443,
                purpose=CertPurpose.CLIENT,
                cert_path=cert_path,
                key_path=key_path)
        finally:
            os.unlink(cert_path)
            os.unlink(key_path)

        self.assertEqual(cert.cert_data, TEST_CLIENT_CERT_PEM)
        self.assertEqual(cert.key_data, TEST_CLIENT_KEY_PEM)

    def test_create_from_files_with_cert_not_found(self) -> None:
        """Testing Certificate.create_from_files with cert not found"""
        message = 'The SSL/TLS certificate was not found.'

        with self.assertRaisesMessage(CertificateNotFoundError, message):
            Certificate.create_from_files(
                hostname='example.com',
                port=443,
                cert_path='/rb-tests-xxx/bad/')

    def test_create_from_files_with_ioerror_cert(self) -> None:
        """Testing Certificate.create_from_files with IOError reading cert"""
        self.spy_on(os.path.exists, op=kgb.SpyOpReturn(True))

        message = (
            r'Error reading SSL/TLS certificate file\. Administrators can '
            r'find details in the Review Board server logs \(error ID '
            r'[a-z0-9-]+\)\.'
        )

        with self.assertRaisesRegex(CertificateStorageError, message):
            Certificate.create_from_files(
                hostname='example.com',
                port=443,
                cert_path='/rb-tests/bad-path/')

    def test_create_from_files_with_invalid_cert(self) -> None:
        """Testing Certificate.create_from_files with invalid cert format"""
        cert_fd, cert_path = tempfile.mkstemp()
        os.write(cert_fd, b'XXX')
        os.close(cert_fd)

        try:
            with self.assertRaises(InvalidCertificateFormatError) as ctx:
                Certificate.create_from_files(
                    hostname='example.com',
                    port=443,
                    cert_path=cert_path)
        finally:
            os.unlink(cert_path)

        self.assertEqual(ctx.exception.data, b'XXX')
        self.assertEqual(ctx.exception.path, cert_path)

    def test_create_from_files_with_client_and_key_not_found(self) -> None:
        """Testing Certificate.create_from_files with purpose=client and
        key not found
        """
        cert_fd, cert_path = tempfile.mkstemp()
        os.write(cert_fd, TEST_CLIENT_CERT_PEM)
        os.close(cert_fd)

        message = 'The SSL/TLS private key was not found.'

        try:
            with self.assertRaisesMessage(CertificateNotFoundError, message):
                Certificate.create_from_files(
                    hostname='example.com',
                    port=443,
                    purpose=CertPurpose.CLIENT,
                    cert_path=cert_path,
                    key_path='/rb-tests-xxx/bad/')
        finally:
            os.unlink(cert_path)

    def test_create_from_files_with_client_and_ioerror_key(self) -> None:
        """Testing Certificate.create_from_files with purpose=client and
        IOError reading key
        """
        cert_fd, cert_path = tempfile.mkstemp()
        os.write(cert_fd, TEST_CLIENT_CERT_PEM)
        os.close(cert_fd)

        self.spy_on(os.path.exists, op=kgb.SpyOpReturn(True))

        message = (
            r'Error reading SSL/TLS private key file\. Administrators can '
            r'find details in the Review Board server logs \(error ID '
            r'[a-z0-9-]+\)\.'
        )

        try:
            with self.assertRaisesRegex(CertificateStorageError, message):
                Certificate.create_from_files(
                    hostname='example.com',
                    port=443,
                    purpose=CertPurpose.CLIENT,
                    cert_path=cert_path,
                    key_path='/rb-tests/bad-path/')
        finally:
            os.unlink(cert_path)

    def test_create_from_files_with_client_and_invalid_key(self) -> None:
        """Testing Certificate.create_from_files with purpose=client and
        invalid key format
        """
        cert_fd, cert_path = tempfile.mkstemp()
        os.write(cert_fd, TEST_CLIENT_CERT_PEM)
        os.close(cert_fd)

        key_fd, key_path = tempfile.mkstemp()
        os.write(key_fd, b'XXX')
        os.close(key_fd)

        try:
            with self.assertRaises(InvalidCertificateFormatError) as ctx:
                Certificate.create_from_files(
                    hostname='example.com',
                    port=443,
                    purpose=CertPurpose.CLIENT,
                    cert_path=cert_path,
                    key_path=key_path)
        finally:
            os.unlink(cert_path)
            os.unlink(key_path)

        self.assertEqual(ctx.exception.data, b'XXX')
        self.assertEqual(ctx.exception.path, key_path)

    @online_only
    def test_create_from_server(self) -> None:
        """Testing Certificate.create_from_server"""
        cert = Certificate.create_from_server(hostname='reviewboard.org',
                                              port=443)

        assert cert is not None
        self.assertAttrsEqual(
            cert,
            {
                'hostname': 'reviewboard.org',
                'port': 443,
            })
        self.assertIsNotNone(cert.cert_data)

    @online_only
    def test_create_from_server_with_bad_cert(self) -> None:
        """Testing Certificate.create_from_server with bad certificate state
        """
        cert = Certificate.create_from_server(hostname='expired.badssl.com',
                                              port=443)

        assert cert is not None
        self.assertAttrsEqual(
            cert,
            {
                'hostname': 'expired.badssl.com',
                'port': 443,
            })
        self.assertIsNotNone(cert.cert_data)

    @online_only
    def test_create_from_server_with_failure(self) -> None:
        """Testing Certificate.create_from_server with failure looking up
        cert
        """
        with self.assertLogs() as cm:
            cert = Certificate.create_from_server(
                hostname='xxx.badhosttld',
                port=443,
            )

        self.assertEqual(len(cm.output), 1)
        self.assertTrue(cm.output[0].startswith(
            'WARNING:reviewboard.certs.cert:Failed to fetch server '
            'certificate from xxx.badhosttld:443:'
        ))

        self.assertIsNone(cert)

    def test_init_with_valid_from_naive(self) -> None:
        """Testing Certificate.__init__ with valid_from as naive datetime"""
        message = 'valid_from must contain a timezone.'

        with self.assertRaisesMessage(ValueError, message):
            Certificate(hostname='example.com',
                        port=443,
                        valid_from=datetime(2023, 7, 14, 7, 50, 30))

    def test_init_with_valid_through_naive(self) -> None:
        """Testing Certificate.__init__ with valid_through as naive datetime"""
        message = 'valid_through must contain a timezone.'

        with self.assertRaisesMessage(ValueError, message):
            Certificate(hostname='example.com',
                        port=443,
                        valid_through=datetime(2023, 7, 14, 7, 50, 30))

    def test_x509_cert(self) -> None:
        """Testing Certificate.x509_cert"""
        cert = Certificate(hostname='example.com',
                           port=443,
                           cert_data=TEST_TRUST_CERT_PEM)

        x509_cert = cert.x509_cert
        assert x509_cert is not None

        name = (
            x509_cert.subject
            .get_attributes_for_oid(NameOID.COMMON_NAME)[0]
            .value
        )

        self.assertEqual(name, 'example.com')

        # This should be cached.
        self.assertIs(cert.x509_cert, x509_cert)

    def test_x509_cert_without_cert_data(self) -> None:
        """Testing Certificate.x509_cert without cert_data"""
        cert = Certificate(hostname='example.com',
                           port=443)

        self.assertIsNone(cert.x509_cert)

    def test_attrs_default(self) -> None:
        """Testing Certificate certificate attribute defaults"""
        cert = Certificate(hostname='example.com',
                           port=443)

        self.assertIsNone(cert.fingerprints)
        self.assertIsNone(cert.issuer)
        self.assertIsNone(cert.subject)
        self.assertEqual(cert.subject_alternative_names, [])
        self.assertIsNone(cert.valid_from)
        self.assertIsNone(cert.valid_through)

    def test_attrs_with_provided(self) -> None:
        """Testing Certificate certificate attributes provided during
        construction
        """
        cert = Certificate(
            hostname='example.com',
            port=443,
            subject='Subject',
            issuer='Issuer',
            fingerprints=CertificateFingerprints(sha1=TEST_SHA1,
                                                 sha256=TEST_SHA256),
            valid_from=datetime(2023, 7, 14, 7, 50, 30, tzinfo=timezone.utc),
            valid_through=datetime(2024, 7, 13, 7, 50, 30,
                                   tzinfo=timezone.utc))

        self.assertEqual(cert.subject, 'Subject')
        self.assertEqual(cert.issuer, 'Issuer')
        self.assertEqual(cert.subject_alternative_names, [])
        self.assertEqual(cert.valid_from,
                         datetime(2023, 7, 14, 7, 50, 30, tzinfo=timezone.utc))
        self.assertEqual(cert.valid_through,
                         datetime(2024, 7, 13, 7, 50, 30, tzinfo=timezone.utc))

        fingerprints = cert.fingerprints
        assert fingerprints is not None

        self.assertEqual(fingerprints.sha1, TEST_SHA1)
        self.assertEqual(fingerprints.sha256, TEST_SHA256)

    def test_attrs_with_x509(self) -> None:
        """Testing Certificate certificate attributes with cert data"""
        cert = Certificate(hostname='example.com',
                           port=443,
                           cert_data=TEST_TRUST_CERT_PEM)

        self.assertEqual(cert.subject, 'example.com')
        self.assertEqual(cert.issuer, 'example.com')
        self.assertEqual(cert.subject_alternative_names, [])
        self.assertEqual(cert.valid_from,
                         datetime(2023, 7, 14, 7, 50, 30, tzinfo=timezone.utc))
        self.assertEqual(cert.valid_through,
                         datetime(2024, 7, 13, 7, 50, 30, tzinfo=timezone.utc))

        fingerprints = cert.fingerprints
        assert fingerprints is not None

        self.assertEqual(fingerprints.sha1, TEST_SHA1)
        self.assertEqual(fingerprints.sha256, TEST_SHA256)

        # This should be cached.
        self.assertIs(cert.fingerprints, fingerprints)

    def test_attr_x509_subject_alt_names(self) -> None:
        """Testing Certificate.subject_alternative_names"""
        cert = Certificate(
            hostname='example.com',
            port=443,
            cert_data=b"""-----BEGIN CERTIFICATE-----
MIIEWzCCAkOgAwIBAgIUH/aJUMZgutxwbM0xXq8DRrfVR+gwDQYJKoZIhvcNAQEL
BQAwFzEVMBMGA1UEAwwMVGVzdCBMREFQIENBMB4XDTI2MDMxNjIzNDM0NVoXDTM2
MDMxMzIzNDM0NVowFDESMBAGA1UEAwwJbG9jYWxob3N0MIIBIjANBgkqhkiG9w0B
AQEFAAOCAQ8AMIIBCgKCAQEAvR432MX+pKoZcpXG7hP2DL6C3YlBggoDJ6YoKz1V
k8dNyX4jWkcYYXbDAMMWd72LQl/LBJeKguaeQpVqv3ViWCgK6aompJDkqyTNqs3l
sbL7ZXzBT+6E/iwP6ChxXEVY8EsBC8kAzBoq1gG1QvtT0qj1P0FpgPKl2T7SxwUG
B+k9gHNwHRRoMiFDVozbB3l6KHUaTro3RmVbAngzgK+2fYv5rXRKhdaKlIU555To
sExOXW9OCyiT7gUX8F/UGp05MhdGoJeB6rj6/Y6ekIubQbzW3JiPWXHuCelO/Iel
VNHUVRacUjXMndXVFx6J+AEmhIdhLK193WZoXV5eAMmd0QIDAQABo4GhMIGeMAkG
A1UdEwQCMAAwDgYDVR0PAQH/BAQDAgWgMBMGA1UdJQQMMAoGCCsGAQUFBwMBMCwG
A1UdEQQlMCOCCWxvY2FsaG9zdIcEfwAAAYcQAAAAAAAAAAAAAAAAAAAAATAdBgNV
HQ4EFgQURPbKJRFrb6AbUB4GuvEWMIKwCFcwHwYDVR0jBBgwFoAUFZ8R0MgmCmxb
G0buLdWs5L+1kT0wDQYJKoZIhvcNAQELBQADggIBACRE4MjIm1hlC5+xiBlyrQ1i
B1TWcRkX9+0UQfI4BMdaIPrEZLz1B8mdRcRAn5NKyfjQKhUsbW7aBXUD/3wsxuy+
Lfgek8xvDoeY4y/L4wkdFl+QfYKgwLsXrzMexHwcCZ3SEvaqFrm55hW0hVcX5FeW
4FARqlGIWp+MSPfGtDHbncGUnk+fRO4EGiZcsakQNdzL7Gn+aS/1swk0A0w6V5bx
+/LkMhkTEZt+fClRAMNW2X169ki+/vDe7pQQinrQ6j+8fDwMvrAD+qpSM91CJLNA
ddMXmrDbvYgDQdQ+YbineuSahoIjgPzE8hJ8rSVl5R+uGSjTgbS5JsrqlA+bP0se
/VzkkIglRm5XmGUHYT4CQIAbbg1u5wQsOrgaymBbFWN08eNhNAak0j0KhmF1PDy6
LML/n+xSjIDwI4llZkog1hv1jmSCW0PJ9ty6UHb6yfZ063spR/SIpMtekGkLllAR
RwPgp5YWsu1t/GiEk0qWlXT6qRxtjYh2zL5SorEAmJtwOYZj5ywuKG4NUzZQ9vqD
0PsulCD8L5we9EAuxYY9D9wQL7qlQLoPnL01FRscsVFHEw6Dv3noqCa0BzjMcU3H
RS9daj4M6YOwBbgGt+YgjLCGsQ7fRGXYo1UebeZDFE6RwLJ0FQvzUQQfuEk7qSq3
pa6+8p+V5VkIunYFLWOw
-----END CERTIFICATE-----""")

        self.assertEqual(cert.subject_alternative_names, [
            'localhost',
            ipaddress.IPv4Address('127.0.0.1'),
            ipaddress.IPv6Address('::1'),
        ])

    def test_is_valid_with_not_expired(self) -> None:
        """Testing Certificate.is_valid with not expired"""
        cert_data = self.build_x509_cert_pem()

        cert = Certificate(hostname='example.com',
                           port=443,
                           cert_data=cert_data)

        self.assertTrue(cert.is_valid)

    def test_is_valid_with_expired(self) -> None:
        """Testing Certificate.is_valid with expired"""
        cert_data = self.build_x509_cert_pem(
            not_valid_before_delta=-timedelta(days=10),
            not_valid_after_delta=-timedelta(days=5))

        cert = Certificate(hostname='example.com',
                           port=443,
                           cert_data=cert_data)

        self.assertFalse(cert.is_valid)

    def test_is_valid_with_not_yet_valid(self) -> None:
        """Testing Certificate.is_valid with not yet valid"""
        cert_data = self.build_x509_cert_pem(
            not_valid_before_delta=timedelta(days=5),
            not_valid_after_delta=timedelta(days=10))

        cert = Certificate(hostname='example.com',
                           port=443,
                           cert_data=cert_data)

        self.assertFalse(cert.is_valid)

    def test_is_wildcard_with_true(self) -> None:
        """Testing Certificate.is_wildcard with wildcard domain"""
        cert = Certificate(hostname='w*.example.com',
                           port=443)

        self.assertTrue(cert.is_wildcard)

    def test_is_wildcard_with_false(self) -> None:
        """Testing Certificate.is_wildcard with non-wildcard domain"""
        cert = Certificate(hostname='www.example.com',
                           port=443)

        self.assertFalse(cert.is_wildcard)

    def test_matches_host_with_equal(self) -> None:
        """Testing Certificate.matches_host with normalized hostnames
        equal
        """
        cert = Certificate(hostname='XXX',
                           port=443)
        cert._subject = 'www.example.COM'
        cert.__dict__['subject_alternative_names'] = [
            'sub.example.net',
        ]

        self.assertTrue(cert.matches_host('WWW.EXAMPLE.COM'))
        self.assertTrue(cert.matches_host('sub.EXAMPLE.net'))

    def test_matches_host_with_bare_hostname_equal(self) -> None:
        """Testing Certificate.matches_host with bare hostname equal"""
        cert = Certificate(hostname='XXX',
                           port=443)
        cert.__dict__['subject_alternative_names'] = [
            'localhost',
        ]

        self.assertTrue(cert.matches_host('localhost'))

    def test_matches_host_with_ipv4_equal(self) -> None:
        """Testing Certificate.matches_host with IPv4 address equal"""
        cert = Certificate(hostname='XXX',
                           port=443)
        cert.__dict__['subject_alternative_names'] = [
            ipaddress.IPv4Address('1.2.3.4'),
        ]

        self.assertTrue(cert.matches_host('1.2.3.4'))

    def test_matches_host_with_ipv6_equal(self) -> None:
        """Testing Certificate.matches_host with IPv6 address equal"""
        cert = Certificate(hostname='XXX',
                           port=443)
        cert.__dict__['subject_alternative_names'] = [
            ipaddress.IPv6Address('2001:db8::1'),
        ]

        self.assertTrue(cert.matches_host('2001:db8::1'))

    def test_matches_host_with_cert_data(self) -> None:
        """Testing Certificate.matches_host with parsed cert data"""
        cert = Certificate(
            hostname='XXX',
            port=443,
            cert_data=b"""-----BEGIN CERTIFICATE-----
MIIDqjCCApKgAwIBAgIUFXRspvQcuDirYjGMSozXtDhIShcwDQYJKoZIhvcNAQEL
BQAwPzELMAkGA1UEBhMCVVMxGjAYBgNVBAoMEUV4YW1wbGUgVGVzdCBDZXJ0MRQw
EgYDVQQDDAtleGFtcGxlLmNvbTAeFw0yNjA0MjAwOTU4MTBaFw0yNzA0MjAxMDAz
MTBaMD8xCzAJBgNVBAYTAlVTMRowGAYDVQQKDBFFeGFtcGxlIFRlc3QgQ2VydDEU
MBIGA1UEAwwLZXhhbXBsZS5jb20wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEK
AoIBAQC/YxvbVPEoAJe9TBja4onX4DH4d8ig7DC2yElO5LDV9qUDGIzZOz1UYJs3
WQtIt0H9Yil4xU18fea2nP3tJFCZv384zuwTzOy6jP38EWVVUhELy88RzmKi/HEm
bcbxO4B1QSTlRrntqpYny0OqZN04zzrpUHz9VyILN0H5X+CoBQ9xLW5t7n0pW9Q+
E59og2eL/BNIQ8syIaLMmCUfhCi8eQqzgOwM8HaEK8TCohyFZ076XzCzrZltVbKH
egCy7M3VXkt2SH8Op72XL8CI6sjlI5qwzw3VfGBpEjvl5boJuXXTLnjLsSU7xRPw
mpzYf3xvZv40L4X1Mqkd3p8PREB3AgMBAAGjgZ0wgZowRgYDVR0RBD8wPYIHZXhh
bXBsZYILZXhhbXBsZS5jb22CDSouZXhhbXBsZS5uZXSHBAECAwSHECABDbgAAAAA
AAAAAAAAAAEwDAYDVR0TAQH/BAIwADAOBgNVHQ8BAf8EBAMCBaAwEwYDVR0lBAww
CgYIKwYBBQUHAwEwHQYDVR0OBBYEFKHmk15f72xt/gnkUqZQ8y8ZiqJXMA0GCSqG
SIb3DQEBCwUAA4IBAQCmtH5xWJq1avwdWDuLC6KN5b8G/JSOmKvPYaMpLijEvJ2p
jlbUYjV5HscC8NMh+21YNIYG7nCE3EZY1+YKGZyo3sjVmZ5pQSRWyaNpA5rTu4v0
WmtGZOsMZFPSGgdBb9pi3ke8oKV1k0gPBnplOc1TDsGRuA+DW2AFHhbbt1oXqJHT
0teI2u9Qu4A+uI6/RxdxZBW5TGw0EKOCTfV7we5FKoM47TCSKB85tTFnRYUVrKQI
2jClbc0ud+V0dS2VDVRSiD3n7E8uHQAHYl1DijMuhkA1Rzs95QoKkTzmjHhvKJ/i
+1czCwcQcFogLXoJel+4GXRcf+Nba9zCn8uX0jJR
-----END CERTIFICATE-----""",
        )

        # These should match.
        self.assertTrue(cert.matches_host('example'))
        self.assertTrue(cert.matches_host('example.com'))
        self.assertTrue(cert.matches_host('Sub.Example.net'))
        self.assertTrue(cert.matches_host('1.2.3.4'))
        self.assertTrue(cert.matches_host('2001:db8::1'))
        self.assertTrue(cert.matches_host('2001:0db8::0001'))

        # These should not.
        self.assertFalse(cert.matches_host('example.net'))
        self.assertFalse(cert.matches_host('1.2.4.5'))
        self.assertFalse(cert.matches_host('2001:db9::1'))

    def test_matches_host_with_not_equal(self) -> None:
        """Testing Certificate.matches_host with hostnames not equal and
        non-wildcard
        """
        cert = Certificate(hostname='XXX',
                           port=443)
        cert._subject = 'example.com'

        self.assertFalse(cert.matches_host('www.example.com'))

    def test_matches_host_with_wildcard(self) -> None:
        """Testing Certificate.matches_host with wildcard for full
        first label
        """
        cert = Certificate(hostname='*.example.com',
                           port=443)
        cert._subject = '*.example.com'
        cert.__dict__['subject_alternative_names'] = [
            '*.example.net',
            '*.sub.example.corp',
        ]

        # These should match.
        self.assertTrue(cert.matches_host('www.example.com'))
        self.assertTrue(cert.matches_host('foobar.example.com'))
        self.assertTrue(cert.matches_host('foobar.example.NET'))
        self.assertTrue(cert.matches_host('xxxyyyzzz.sub.example.corp'))

        # These should not.
        self.assertFalse(cert.matches_host('example.com'))
        self.assertFalse(cert.matches_host('sub.www.example.com'))
        self.assertFalse(cert.matches_host('sub.www.example.net'))

    def test_write_cert_file(self) -> None:
        """Testing Certificate.write_cert_file"""
        cert = Certificate(hostname='example.com',
                           port=443,
                           cert_data=TEST_TRUST_CERT_PEM)

        fd, path = tempfile.mkstemp()
        os.close(fd)

        try:
            cert.write_cert_file(path)

            with open(path, 'rb') as fp:
                self.assertEqual(fp.read(), TEST_TRUST_CERT_PEM)
        finally:
            os.unlink(path)

    def test_write_cert_file_with_ioerror(self) -> None:
        """Testing Certificate.write_cert_file with IOError"""
        cert = Certificate(hostname='example.com',
                           port=443,
                           cert_data=TEST_TRUST_CERT_PEM)

        message = (
            r'Error writing SSL/TLS certificate file\. Administrators can '
            r'find details in the Review Board server logs \(error ID '
            r'[a-z0-9-]+\)\.'
        )

        with self.assertRaisesRegex(CertificateStorageError, message):
            cert.write_cert_file('/rb-tests-xxx/bad/')

    def test_write_key_file(self) -> None:
        """Testing Certificate.write_key_file"""
        cert = Certificate(hostname='example.com',
                           port=443,
                           purpose=CertPurpose.CLIENT,
                           cert_data=TEST_CLIENT_CERT_PEM,
                           key_data=TEST_CLIENT_KEY_PEM)

        fd, path = tempfile.mkstemp()
        os.close(fd)

        try:
            cert.write_key_file(path)

            with open(path, 'rb') as fp:
                self.assertEqual(fp.read(), TEST_CLIENT_KEY_PEM)
        finally:
            os.unlink(path)

    def test_write_key_file_with_ioerror(self) -> None:
        """Testing Certificate.write_key_file with IOError"""
        cert = Certificate(hostname='example.com',
                           port=443,
                           purpose=CertPurpose.CLIENT,
                           cert_data=TEST_CLIENT_CERT_PEM,
                           key_data=TEST_CLIENT_KEY_PEM)

        message = (
            r'Error writing SSL/TLS private key file\. Administrators can '
            r'find details in the Review Board server logs \(error ID '
            r'[a-z0-9-]+\)\.'
        )

        with self.assertRaisesRegex(CertificateStorageError, message):
            cert.write_key_file('/rb-tests-xxx/bad/')

    def test_repr(self) -> None:
        """Testing Certificate.__repr__"""
        cert = Certificate(hostname='example.com',
                           port=443)

        self.assertEqual(
            repr(cert),
            "<Certificate(hostname='example.com', port=443, "
            "purpose=CertPurpose.TRUST, fingerprints=None)>")

    def test_repr_with_fingerprints(self) -> None:
        """Testing Certificate.__repr__ with fingerprints"""
        cert = Certificate(
            hostname='example.com',
            port=443,
            fingerprints=CertificateFingerprints(sha1=TEST_SHA1))

        self.assertEqual(
            repr(cert),
            "<Certificate(hostname='example.com', port=443, "
            "purpose=CertPurpose.TRUST, "
            "fingerprints=<CertificateFingerprints("
            "sha1='F2:35:0F:BB:34:40:84:78:8B:20:1D:40:B1:4A:17:0C:DE:36:2F:"
            "D5', sha256=None)>)>")

    def test_to_json(self) -> None:
        """Testing Certificate.to_json"""
        fingerprints = CertificateFingerprints(sha1=TEST_SHA1,
                                               sha256=TEST_SHA256)

        cert = Certificate(
            hostname='example.com',
            port=443,
            subject='Subject',
            issuer='Issuer',
            fingerprints=CertificateFingerprints(sha1=TEST_SHA1,
                                                 sha256=TEST_SHA256),
            valid_from=datetime(2023, 7, 14, 7, 50, 30, tzinfo=timezone.utc),
            valid_through=datetime(2024, 7, 13, 7, 50, 30,
                                   tzinfo=timezone.utc))

        json_data = cert.to_json()

        self.assertEqual(
            json_data,
            {
                'fingerprints': fingerprints,
                'hostname': 'example.com',
                'issuer': 'Issuer',
                'port': 443,
                'purpose': 'trust',
                'subject': 'Subject',
                'valid_from': datetime(2023, 7, 14, 7, 50, 30,
                                       tzinfo=timezone.utc),
                'valid_through': datetime(2024, 7, 13, 7, 50, 30,
                                          tzinfo=timezone.utc),
            })

        # Verify we used encodeable data.
        self.assertJSONEqual(
            json.dumps(json_data, cls=DjbletsJSONEncoder),
            {
                'fingerprints': {
                    'sha1': (
                        'F2:35:0F:BB:34:40:84:78:8B:20:1D:40:B1:4A:17:0C:DE:'
                        '36:2F:D5'
                    ),
                    'sha256': (
                        '79:19:70:AE:A6:1B:EB:BC:35:7C:B8:54:B1:6A:AD:79:FF:'
                        'F7:28:69:02:5E:C3:6F:B3:C2:B4:FD:84:66:DF:8F'
                    ),
                },
                'hostname': 'example.com',
                'issuer': 'Issuer',
                'port': 443,
                'purpose': 'trust',
                'subject': 'Subject',
                'valid_from': '2023-07-14T07:50:30Z',
                'valid_through': '2024-07-13T07:50:30Z',
            })



"""API resource for managing a repository's group ACL.

Version Added:
    4.0.11
"""

from __future__ import unicode_literals

from django.core.exceptions import ObjectDoesNotExist
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_request_fields,
                                       webapi_response_errors)
from djblets.webapi.errors import (DOES_NOT_EXIST, NOT_LOGGED_IN,
                                   PERMISSION_DENIED)
from djblets.webapi.fields import StringFieldType

from reviewboard.reviews.models import Group
from reviewboard.webapi.decorators import webapi_check_local_site
from reviewboard.webapi.errors import INVALID_GROUP
from reviewboard.webapi.resources import resources
from reviewboard.webapi.resources.review_group import ReviewGroupResource


class RepositoryGroupResource(ReviewGroupResource):
    """Provides information on groups that are allowed access to a repository.

    Version Added:
        4.0.11
    """

    name = 'repository_group'
    item_result_key = 'group'
    list_result_key = 'groups'
    uri_name = 'groups'
    mimetype_list_resource_name = 'repository-groups'
    mimetype_item_resource_name = 'repository-group'

    # We don't want any group child resources to be available under this
    # resource, as they will have the wrong URLs, and do not make sense as
    # sub-resources. We will be serializing a link to the authoritative group
    # resource and children can be found from there.
    item_child_resources = []

    allowed_methods = ('GET', 'POST', 'DELETE')

    policy_id = 'repository_group'

    def get_queryset(self, request, *args, **kwargs):
        """Return a queryset for the repository users.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments passed through to the parent resource.

            **kwargs (dict):
                Keyword arguments passed through to the parent resource.

        Returns:
            django.db.models.query.QuerySet:
            The queryset for the users.
        """
        try:
            repository = resources.repository.get_object(request, *args,
                                                         **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        return repository.review_groups.all()

    def get_href_parent_ids(self, obj, **kwargs):
        """Return the href parent IDs for the object.

        Args:
            obj (reviewboard.reviews.models.Group):
                The review group.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict:
            The parent IDs to be used to determine the href of the resource.
        """
        # Since we do not have a direct link to the model parent (the
        # Repository.review_groups field is a many-to-many field so we cannot
        # use it because the reverse relation is not unique), we have to
        # manually generate the parent IDs from the parent resource.
        parent_id_key = self._parent_resource.uri_object_key

        return {
            parent_id_key: kwargs[parent_id_key],
        }

    def get_related_links(self, obj=None, request=None, *args, **kwargs):
        """Return the related links for the resource.

        Args:
            obj (reviewboard.reviews.models.Group, optional):
                The group for which links are being generated.

            request (django.http.HttpRequest, optional):
                The current HTTP request.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict:
            The related links for the resource.
        """
        links = super(RepositoryGroupResource, self).get_related_links(
            obj, request=request, *args, **kwargs)

        # We only want the 'group' link when this is an item resource.
        if self.uri_object_key in kwargs:
            group_name = kwargs[self.uri_object_key]
            links['group'] = {
                'href': resources.review_group.get_item_url(
                    group_name=group_name),
                'method': 'GET',
            }

        return links

    def get_serializer_for_object(self, obj):
        """Return the serializer for an object.

        If the object is a :py:class:`~reviewboard.reviews.models.Group`
        instance, we will serialize it (instead of the
        :py:class:`~reviewboard.webapi.resources.review_group.ReviewGroupResource`
        resource so that the links will be correct. Otherwise, the POST and
        DELETE links will be for the actual user instead of for this resource.

        Args:
            obj (django.db.models.base.Model):
                The model being serialized.

        Returns:
            djblets.webapi.resources.base.WebAPIResource:
            The resource that should be used to serialize the object.
        """
        if isinstance(obj, Group):
            return self

        return super(RepositoryGroupResource, self).get_serializer_for_object(
            obj)

    def has_access_permissions(self, request, group, *args, **kwargs):
        """Return whether the item resource can be accessed.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            group (reviewboard.reviews.models.group.Group, unused):
                The group in the resource item URL. This is unused because we
                only care about repository-level access here.

            *args (tuple):
                Positional arguments to pass to the parent resource.

            **kwargs (dict):
                Keyword arguments to pass to the parent resource.

        Returns:
            bool:
            Whether the current user can access the item resource.
        """
        repository = resources.repository.get_object(request, *args, **kwargs)
        return repository.is_mutable_by(request.user)

    def has_list_access_permissions(self, request, *args, **kwargs):
        """Return whether the list resource can be accessed.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments to pass to the parent resource.

            **kwargs (dict):
                Keyword arguments to pass to the parent resource.

        Returns:
            bool:
            Whether the current user can access the list resource.
        """
        repository = resources.repository.get_object(request, *args, **kwargs)
        return repository.is_mutable_by(request.user)

    def has_modify_permissions(self, request, *args, **kwargs):
        """Return whether the resource can be modified.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments to pass to the parent resource.

            **kwargs (dict):
                Keyword arguments to pass to the parent resource.

        Returns:
            bool:
            Whether the current user can modify the resource.
        """
        repository = resources.repository.get_object(request, *args, **kwargs)
        return repository.is_mutable_by(request.user)

    def has_delete_permissions(self, request, *args, **kwargs):
        """Return whether the resource can be deleted.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments to pass to the parent resource.

            **kwargs (dict):
                Keyword arguments to pass to the parent resource.

        Returns:
            bool:
            Whether the current user can delete the resource.
        """
        repository = resources.repository.get_object(request, *args, **kwargs)
        return repository.is_mutable_by(request.user)

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, INVALID_GROUP, NOT_LOGGED_IN,
                            PERMISSION_DENIED)
    @webapi_request_fields(required={
        'group_name': {
            'type': StringFieldType,
            'description': 'The group to add to the repository ACL.',
        },
    })
    def create(self, request, group_name, local_site_name=None,
               *args, **kwargs):
        """Adds a group to the repository ACL."""
        repo_resource = resources.repository

        try:
            repository = repo_resource.get_object(
                request, local_site_name=local_site_name, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not repo_resource.has_modify_permissions(request, repository):
            return self.get_no_access_error(request)

        local_site = self._get_local_site(local_site_name)

        try:
            if local_site:
                group = local_site.groups.get(name=group_name,
                                              invite_only=True)
            else:
                group = Group.objects.get(name=group_name,
                                          invite_only=True)
        except ObjectDoesNotExist:
            return INVALID_GROUP

        repository.review_groups.add(group)

        return 201, {
            self.item_result_key: group,
        }

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, NOT_LOGGED_IN, PERMISSION_DENIED)
    def delete(self, request, *args, **kwargs):
        """Removes a group from the repository ACL."""
        repo_resource = resources.repository

        try:
            repository = repo_resource.get_object(request, *args, **kwargs)
            group = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not repo_resource.has_modify_permissions(request, repository):
            return self.get_no_access_error(request)

        repository.review_groups.remove(group)

        return 204, {}


repository_group_resource = RepositoryGroupResource()
